import pytest

from cloudpathlib import GSPath
from cloudpathlib.local import LocalGSPath


@pytest.mark.parametrize("path_class", [GSPath, LocalGSPath])
def test_gspath_properties(path_class):
    p = path_class("gs://mybucket")
    assert p.blob == ""
    assert p.bucket == "mybucket"

    p2 = path_class("gs://mybucket/")
    assert p2.blob == ""
    assert p2.bucket == "mybucket"



import os
from pathlib import Path, PurePosixPath
import shutil

from azure.storage.blob import BlobServiceClient
import boto3
import botocore
from dotenv import find_dotenv, load_dotenv
from google.cloud import storage as google_storage
from pytest_cases import fixture, fixture_union
from shortuuid import uuid

from cloudpathlib import AzureBlobClient, AzureBlobPath, GSClient, GSPath, S3Client, S3Path
from cloudpathlib.cloudpath import implementation_registry
from cloudpathlib.local import (
    local_azure_blob_implementation,
    LocalAzureBlobClient,
    LocalAzureBlobPath,
    local_gs_implementation,
    LocalGSClient,
    LocalGSPath,
    local_s3_implementation,
    LocalS3Client,
    LocalS3Path,
)
import cloudpathlib.azure.azblobclient
import cloudpathlib.s3.s3client
from .mock_clients.mock_azureblob import mocked_client_class_factory
from .mock_clients.mock_gs import mocked_client_class_factory as mocked_gsclient_class_factory
from .mock_clients.mock_s3 import mocked_session_class_factory


if os.getenv("USE_LIVE_CLOUD") == "1":
    load_dotenv(find_dotenv())


SESSION_UUID = uuid()

# ignore these files when uploading test assets
UPLOAD_IGNORE_LIST = [
    ".DS_Store",  # macOS cruft
]


@fixture()
def assets_dir() -> Path:
    """Path to test assets directory."""
    return Path(__file__).parent / "assets"


class CloudProviderTestRig:
    """Class that holds together the components needed to test a cloud implementation."""

    def __init__(
        self,
        path_class: type,
        client_class: type,
        drive: str = "drive",
        test_dir: str = "",
        live_server: bool = False,
    ):
        """
        Args:
            path_class (type): CloudPath subclass
            client_class (type): Client subclass
        """
        self.path_class = path_class
        self.client_class = client_class
        self.drive = drive
        self.test_dir = test_dir
        self.live_server = live_server  # if the server is a live server

    @property
    def cloud_prefix(self):
        return self.path_class.cloud_prefix

    def create_cloud_path(self, path: str):
        """CloudPath constructor that appends cloud prefix. Use this to instantiate
        cloud path instances with generic paths. Includes drive and root test_dir already."""
        return self.path_class(
            cloud_path=f"{self.path_class.cloud_prefix}{self.drive}/{self.test_dir}/{path}"
        )


def create_test_dir_name(request) -> str:
    """Generates unique test directory name using test module and test function names."""
    module_name = request.module.__name__.rpartition(".")[-1]
    function_name = request.function.__name__
    test_dir = f"{SESSION_UUID}-{module_name}-{function_name}"
    print("Test directory name is:", test_dir)
    return test_dir


@fixture()
def azure_rig(request, monkeypatch, assets_dir):
    drive = os.getenv("LIVE_AZURE_CONTAINER", "container")
    test_dir = create_test_dir_name(request)

    live_server = os.getenv("USE_LIVE_CLOUD") == "1"

    if live_server:
        # Set up test assets
        blob_service_client = BlobServiceClient.from_connection_string(
            os.getenv("AZURE_STORAGE_CONNECTION_STRING")
        )
        test_files = [
            f for f in assets_dir.glob("**/*") if f.is_file() and f.name not in UPLOAD_IGNORE_LIST
        ]
        for test_file in test_files:
            blob_client = blob_service_client.get_blob_client(
                container=drive,
                blob=str(f"{test_dir}/{PurePosixPath(test_file.relative_to(assets_dir))}"),
            )
            blob_client.upload_blob(test_file.read_bytes(), overwrite=True)
    else:
        monkeypatch.setenv("AZURE_STORAGE_CONNECTION_STRING", "")
        # Mock cloud SDK
        monkeypatch.setattr(
            cloudpathlib.azure.azblobclient,
            "BlobServiceClient",
            mocked_client_class_factory(test_dir),
        )

    rig = CloudProviderTestRig(
        path_class=AzureBlobPath,
        client_class=AzureBlobClient,
        drive=drive,
        test_dir=test_dir,
        live_server=live_server,
    )

    rig.client_class().set_as_default_client()  # set default client

    yield rig

    rig.client_class._default_client = None  # reset default client

    if live_server:
        # Clean up test dir
        container_client = blob_service_client.get_container_client(drive)
        to_delete = container_client.list_blobs(name_starts_with=test_dir)
        container_client.delete_blobs(*to_delete)


@fixture()
def gs_rig(request, monkeypatch, assets_dir):
    drive = os.getenv("LIVE_GS_BUCKET", "bucket")
    test_dir = create_test_dir_name(request)

    live_server = os.getenv("USE_LIVE_CLOUD") == "1"

    if live_server:
        # Set up test assets
        bucket = google_storage.Client().bucket(drive)
        test_files = [
            f for f in assets_dir.glob("**/*") if f.is_file() and f.name not in UPLOAD_IGNORE_LIST
        ]
        for test_file in test_files:
            blob = google_storage.Blob(
                str(f"{test_dir}/{PurePosixPath(test_file.relative_to(assets_dir))}"),
                bucket,
            )
            blob.upload_from_filename(str(test_file))
    else:
        # Mock cloud SDK
        monkeypatch.setattr(
            cloudpathlib.gs.gsclient,
            "StorageClient",
            mocked_gsclient_class_factory(test_dir),
        )

    rig = CloudProviderTestRig(
        path_class=GSPath,
        client_class=GSClient,
        drive=drive,
        test_dir=test_dir,
        live_server=live_server,
    )

    rig.client_class().set_as_default_client()  # set default client

    yield rig

    rig.client_class._default_client = None  # reset default client

    if live_server:
        # Clean up test dir
        for blob in bucket.list_blobs(prefix=test_dir):
            blob.delete()


@fixture()
def s3_rig(request, monkeypatch, assets_dir):
    drive = os.getenv("LIVE_S3_BUCKET", "bucket")
    test_dir = create_test_dir_name(request)

    live_server = os.getenv("USE_LIVE_CLOUD") == "1"

    if live_server:
        # Set up test assets
        session = boto3.Session()  # Fresh session to ensure isolation
        bucket = session.resource("s3").Bucket(drive)
        test_files = [
            f for f in assets_dir.glob("**/*") if f.is_file() and f.name not in UPLOAD_IGNORE_LIST
        ]
        for test_file in test_files:
            bucket.upload_file(
                str(test_file),
                str(f"{test_dir}/{PurePosixPath(test_file.relative_to(assets_dir))}"),
            )
    else:
        # Mock cloud SDK
        monkeypatch.setattr(
            cloudpathlib.s3.s3client,
            "Session",
            mocked_session_class_factory(test_dir),
        )

    rig = CloudProviderTestRig(
        path_class=S3Path,
        client_class=S3Client,
        drive=drive,
        test_dir=test_dir,
        live_server=live_server,
    )

    rig.client_class().set_as_default_client()  # set default client

    yield rig

    rig.client_class._default_client = None  # reset default client

    if live_server:
        # Clean up test dir
        bucket.objects.filter(Prefix=test_dir).delete()


@fixture()
def custom_s3_rig(request, monkeypatch, assets_dir):
    """
    Custom S3 rig used to test the integrations with non-AWS S3-compatible object storages like
        - MinIO (https://min.io/)
        - CEPH  (https://ceph.io/ceph-storage/object-storage/)
        - others
    """
    drive = os.getenv("CUSTOM_S3_BUCKET", "bucket")
    test_dir = create_test_dir_name(request)
    custom_endpoint_url = os.getenv("CUSTOM_S3_ENDPOINT", "https://s3.us-west-1.drivendatabws.com")

    live_server = os.getenv("USE_LIVE_CLOUD") == "1"

    if live_server:
        monkeypatch.setenv("AWS_ACCESS_KEY_ID", os.getenv("CUSTOM_S3_KEY_ID"))
        monkeypatch.setenv("AWS_SECRET_ACCESS_KEY", os.getenv("CUSTOM_S3_SECRET_KEY"))

        # Upload test assets
        session = boto3.Session()  # Fresh session to ensure isolation from AWS S3 auth
        s3 = session.resource("s3", endpoint_url=custom_endpoint_url)

        # idempotent and our test server on heroku only has ephemeral storage
        # so we need to try to create each time
        try:
            s3.meta.client.head_bucket(Bucket=drive)
        except botocore.exceptions.ClientError:
            s3.create_bucket(Bucket=drive)

        bucket = s3.Bucket(drive)

        test_files = [
            f for f in assets_dir.glob("**/*") if f.is_file() and f.name not in UPLOAD_IGNORE_LIST
        ]
        for test_file in test_files:
            bucket.upload_file(
                str(test_file),
                str(f"{test_dir}/{PurePosixPath(test_file.relative_to(assets_dir))}"),
            )
    else:
        # Mock cloud SDK
        monkeypatch.setattr(
            cloudpathlib.s3.s3client,
            "Session",
            mocked_session_class_factory(test_dir),
        )

    rig = CloudProviderTestRig(
        path_class=S3Path,
        client_class=S3Client,
        drive=drive,
        test_dir=test_dir,
        live_server=live_server,
    )

    rig.client_class(
        endpoint_url=custom_endpoint_url
    ).set_as_default_client()  # set default client

    # add flag for custom_s3 rig to skip some tests
    rig.is_custom_s3 = True

    yield rig

    rig.client_class._default_client = None  # reset default client

    if live_server:
        bucket.objects.filter(Prefix=test_dir).delete()


@fixture()
def local_azure_rig(request, monkeypatch, assets_dir):
    drive = os.getenv("LIVE_AZURE_CONTAINER", "container")
    test_dir = create_test_dir_name(request)

    # copy test assets
    shutil.copytree(assets_dir, LocalAzureBlobClient.get_default_storage_dir() / drive / test_dir)

    monkeypatch.setitem(implementation_registry, "azure", local_azure_blob_implementation)

    rig = CloudProviderTestRig(
        path_class=LocalAzureBlobPath,
        client_class=LocalAzureBlobClient,
        drive=drive,
        test_dir=test_dir,
    )

    monkeypatch.setenv("AZURE_STORAGE_CONNECTION_STRING", "")
    rig.client_class().set_as_default_client()  # set default client

    yield rig

    rig.client_class._default_client = None  # reset default client
    rig.client_class.reset_default_storage_dir()  # reset local storage directory


@fixture()
def local_gs_rig(request, monkeypatch, assets_dir):
    drive = os.getenv("LIVE_GS_BUCKET", "bucket")
    test_dir = create_test_dir_name(request)

    # copy test assets
    shutil.copytree(assets_dir, LocalGSClient.get_default_storage_dir() / drive / test_dir)

    monkeypatch.setitem(implementation_registry, "gs", local_gs_implementation)

    rig = CloudProviderTestRig(
        path_class=LocalGSPath,
        client_class=LocalGSClient,
        drive=drive,
        test_dir=test_dir,
    )

    rig.client_class().set_as_default_client()  # set default client

    yield rig

    rig.client_class._default_client = None  # reset default client
    rig.client_class.reset_default_storage_dir()  # reset local storage directory


@fixture()
def local_s3_rig(request, monkeypatch, assets_dir):
    drive = os.getenv("LIVE_S3_BUCKET", "bucket")
    test_dir = create_test_dir_name(request)

    # copy test assets
    shutil.copytree(assets_dir, LocalS3Client.get_default_storage_dir() / drive / test_dir)

    monkeypatch.setitem(implementation_registry, "s3", local_s3_implementation)

    rig = CloudProviderTestRig(
        path_class=LocalS3Path,
        client_class=LocalS3Client,
        drive=drive,
        test_dir=test_dir,
    )

    rig.client_class().set_as_default_client()  # set default client

    yield rig

    rig.client_class._default_client = None  # reset default client
    rig.client_class.reset_default_storage_dir()  # reset local storage directory


rig = fixture_union(
    "rig",
    [
        azure_rig,
        gs_rig,
        s3_rig,
        custom_s3_rig,
        local_azure_rig,
        local_s3_rig,
        local_gs_rig,
    ],
)

# run some s3-specific tests on custom s3 (ceph, minio, etc.) and aws s3
s3_like_rig = fixture_union(
    "s3_like_rig",
    [
        s3_rig,
        custom_s3_rig,
    ],
)



from pathlib import Path, PosixPath, WindowsPath

import pytest

from cloudpathlib.anypath import AnyPath
from cloudpathlib.cloudpath import CloudPath
from cloudpathlib.exceptions import AnyPathTypeError


def test_anypath_path():
    path = Path("a/b/c")
    assert AnyPath(path) == path
    assert AnyPath(str(path)) == path

    assert isinstance(path, AnyPath)
    assert not isinstance(str(path), AnyPath)

    assert issubclass(Path, AnyPath)
    assert issubclass(PosixPath, AnyPath)
    assert issubclass(WindowsPath, AnyPath)
    assert not issubclass(str, AnyPath)


def test_anypath_cloudpath(rig):
    cloudpath = rig.create_cloud_path("a/b/c")
    assert AnyPath(cloudpath) == cloudpath
    assert AnyPath(str(cloudpath)) == cloudpath

    assert isinstance(cloudpath, AnyPath)
    assert not isinstance(str(cloudpath), AnyPath)

    assert issubclass(CloudPath, AnyPath)
    assert issubclass(rig.path_class, AnyPath)


def test_anypath_bad_input():
    with pytest.raises(AnyPathTypeError):
        AnyPath(0)



from pathlib import PurePosixPath

from cloudpathlib.cloudpath import _resolve


def test_resolve():
    assert _resolve(PurePosixPath("/a/b/c")) == "/a/b/c"
    assert _resolve(PurePosixPath("/a/b/c/")) == "/a/b/c"
    assert _resolve(PurePosixPath("/a/b/c//")) == "/a/b/c"
    assert _resolve(PurePosixPath("//a/b/c//")) == "/a/b/c"
    assert _resolve(PurePosixPath("a/b/c")) == "/a/b/c"
    assert _resolve(PurePosixPath("a/b/c/")) == "/a/b/c"
    assert _resolve(PurePosixPath("a//b/c")) == "/a/b/c"
    assert _resolve(PurePosixPath("a////b/c")) == "/a/b/c"
    assert _resolve(PurePosixPath("a/b/./c")) == "/a/b/c"
    assert _resolve(PurePosixPath("a/b/../c")) == "/a/c"
    assert _resolve(PurePosixPath("a/b/../../c")) == "/c"
    assert _resolve(PurePosixPath("")) == "/"



from pathlib import Path
from time import sleep

import pytest

from cloudpathlib.local import LocalGSPath, LocalS3Path, LocalS3Client
from cloudpathlib.exceptions import (
    CloudPathFileExistsError,
    CloudPathNotADirectoryError,
    OverwriteNewerCloudError,
)


@pytest.fixture
def upload_assets_dir(tmpdir):
    tmp_assets = tmpdir.mkdir("test_upload_from_dir")
    p = Path(tmp_assets)
    (p / "upload_1.txt").write_text("Hello from 1")
    (p / "upload_2.txt").write_text("Hello from 2")

    sub = p / "subdir"
    sub.mkdir(parents=True)
    (sub / "sub_upload_1.txt").write_text("Hello from sub 1")
    (sub / "sub_upload_2.txt").write_text("Hello from sub 2")

    yield p


def assert_mirrored(cloud_path, local_path, check_no_extra=True):
    # file exists and is file
    if local_path.is_file():
        assert cloud_path.exists()
        assert cloud_path.is_file()
    else:
        # all local files exist on cloud
        for lp in local_path.iterdir():
            assert_mirrored(cloud_path / lp.name, lp)

        # no extra files on cloud
        if check_no_extra:
            assert len(list(local_path.glob("**/*"))) == len(list(cloud_path.glob("**/*")))

    return True


def test_upload_from_file(rig, upload_assets_dir):
    to_upload = upload_assets_dir / "upload_1.txt"

    # to file, file does not exists
    p = rig.create_cloud_path("upload_test.txt")
    assert not p.exists()

    p.upload_from(to_upload)
    assert p.exists()
    assert p.read_text() == "Hello from 1"

    # to file, file exists
    to_upload_2 = upload_assets_dir / "upload_2.txt"
    sleep(1.5)
    to_upload_2.touch()  # make sure local is newer
    p.upload_from(to_upload_2)
    assert p.exists()
    assert p.read_text() == "Hello from 2"

    # to file, file exists and is newer
    p.touch()
    with pytest.raises(OverwriteNewerCloudError):
        p.upload_from(upload_assets_dir / "upload_1.txt")

    # to file, file exists and is newer; overwrite
    p.touch()
    sleep(1.5)
    p.upload_from(upload_assets_dir / "upload_1.txt", force_overwrite_to_cloud=True)
    assert p.exists()
    assert p.read_text() == "Hello from 1"

    # to dir, dir exists
    p = rig.create_cloud_path("dir_0")  # created by fixtures
    assert p.exists()
    p.upload_from(upload_assets_dir / "upload_1.txt")
    assert (p / "upload_1.txt").exists()
    assert (p / "upload_1.txt").read_text() == "Hello from 1"


def test_upload_from_dir(rig, upload_assets_dir):
    # to dir, dir does not exists
    p = rig.create_cloud_path("upload_test_dir")
    assert not p.exists()

    p.upload_from(upload_assets_dir)
    assert assert_mirrored(p, upload_assets_dir)

    # to dir, dir exists
    p2 = rig.create_cloud_path("dir_0")  # created by fixtures
    assert p2.exists()

    p2.upload_from(upload_assets_dir)
    assert assert_mirrored(p2, upload_assets_dir, check_no_extra=False)

    # a newer file exists on cloud
    sleep(1)
    (p / "upload_1.txt").touch()
    with pytest.raises(OverwriteNewerCloudError):
        p.upload_from(upload_assets_dir)

    # force overwrite
    (p / "upload_1.txt").touch()
    (p / "upload_2.txt").unlink()
    p.upload_from(upload_assets_dir, force_overwrite_to_cloud=True)
    assert assert_mirrored(p, upload_assets_dir)


def test_copy(rig, upload_assets_dir, tmpdir):
    to_upload = upload_assets_dir / "upload_1.txt"
    p = rig.create_cloud_path("upload_test.txt")
    assert not p.exists()
    p.upload_from(to_upload)
    assert p.exists()

    # cloud to local dir
    dst = Path(tmpdir.mkdir("test_copy_to_local"))
    out_file = p.copy(dst)
    assert out_file.exists()
    assert out_file.read_text() == "Hello from 1"
    out_file.unlink()

    p.copy(str(out_file))
    assert out_file.exists()
    assert out_file.read_text() == "Hello from 1"

    # cloud to local file
    p.copy(dst / "file.txt")

    # cloud to cloud -> make sure no local cache
    p_new = p.copy(p.parent / "new_upload_1.txt")
    assert p_new.exists()
    assert not p_new._local.exists()  # cache should never have been downloaded
    assert not p._local.exists()  # cache should never have been downloaded
    assert p_new.read_text() == "Hello from 1"

    # cloud to cloud path as string
    cloud_dest = str(p.parent / "new_upload_0.txt")
    p_new = p.copy(cloud_dest)
    assert p_new.exists()
    assert p_new.read_text() == "Hello from 1"

    # cloud to cloud directory
    cloud_dest = rig.create_cloud_path("dir_1")  # created by fixtures
    p_new = p.copy(cloud_dest)
    assert str(p_new) == str(p_new.parent / p.name)  # file created
    assert p_new.exists()
    assert p_new.read_text() == "Hello from 1"

    # cloud to cloud overwrite
    p_new.touch()
    with pytest.raises(OverwriteNewerCloudError):
        p_new = p.copy(p_new)

    p_new = p.copy(p_new, force_overwrite_to_cloud=True)
    assert p_new.exists()

    # cloud to other cloud
    p2 = rig.create_cloud_path("dir_0/file0_0.txt")
    other = (
        LocalS3Path("s3://fake-bucket/new_other.txt")
        if not isinstance(p2.client, LocalS3Client)
        else LocalGSPath("gs://fake-bucket/new_other.txt")
    )
    assert not other.exists()

    assert not p2._local.exists()  # not in cache
    p2.copy(other)  # forces download + reupload
    assert p2._local.exists()  # in cache
    assert other.exists()
    assert other.read_text() == p2.read_text()

    other.unlink()

    # cloud to other cloud dir
    other_dir = (
        LocalS3Path("s3://fake-bucket/new_other")
        if not isinstance(p2.client, LocalS3Client)
        else LocalGSPath("gs://fake-bucket/new_other")
    )
    (other_dir / "file.txt").write_text("i am a file")  # ensure other_dir exists
    assert other_dir.exists()
    assert not (other_dir / p2.name).exists()

    p2.copy(other_dir)
    assert (other_dir / p2.name).exists()
    assert (other_dir / p2.name).read_text() == p2.read_text()
    (other_dir / p2.name).unlink()

    # cloud dir raises
    cloud_dir = rig.create_cloud_path("dir_1")  # created by fixtures
    with pytest.raises(ValueError) as e:
        p_new = cloud_dir.copy(Path(tmpdir.mkdir("test_copy_dir_fails")))
        assert "use the method copytree" in str(e)


def test_copytree(rig, tmpdir):
    # cloud file raises
    with pytest.raises(CloudPathNotADirectoryError):
        p = rig.create_cloud_path("dir_0/file0_0.txt")
        local_out = Path(tmpdir.mkdir("copytree_fail_on_file"))
        p.copytree(local_out)

    with pytest.raises(CloudPathFileExistsError):
        p = rig.create_cloud_path("dir_0")
        p_out = rig.create_cloud_path("dir_0/file0_0.txt")
        p.copytree(p_out)

    # cloud dir to local dir that exists
    p = rig.create_cloud_path("dir_1")
    local_out = Path(tmpdir.mkdir("copytree_from_cloud"))
    p.copytree(local_out)
    assert assert_mirrored(p, local_out)

    # str version of path
    local_out = Path(tmpdir.mkdir("copytree_to_str_path"))
    p.copytree(str(local_out))
    assert assert_mirrored(p, local_out)

    # cloud dir to local dir that does not exist
    local_out = local_out / "new_folder"
    p.copytree(local_out)
    assert assert_mirrored(p, local_out)

    # cloud dir to cloud dir that does not exist
    p2 = rig.create_cloud_path("new_dir")
    p.copytree(p2)
    assert assert_mirrored(p2, p)

    # cloud dir to cloud dir that exists
    p2 = rig.create_cloud_path("new_dir2")
    (p2 / "existing_file.txt").write_text("asdf")  # ensures p2 exists
    p.copytree(p2)
    assert assert_mirrored(p2, p, check_no_extra=False)

    (p / "new_file.txt").write_text("hello!")  # add file so we can assert mirror
    with pytest.raises(OverwriteNewerCloudError):
        p.copytree(p2)

    p.copytree(p2, force_overwrite_to_cloud=True)
    assert assert_mirrored(p2, p, check_no_extra=False)



import os

import pytest

from cloudpathlib import AzureBlobClient, AzureBlobPath
from cloudpathlib.exceptions import MissingCredentialsError
from cloudpathlib.local import LocalAzureBlobClient, LocalAzureBlobPath


@pytest.mark.parametrize("path_class", [AzureBlobPath, LocalAzureBlobPath])
def test_azureblobpath_properties(path_class, monkeypatch):
    if not os.getenv("AZURE_STORAGE_CONNECTION_STRING"):
        monkeypatch.setenv("AZURE_STORAGE_CONNECTION_STRING", "AccountName=fake;AccountKey=fake2;")

    p = path_class("az://mycontainer")
    assert p.blob == ""
    assert p.container == "mycontainer"

    p2 = path_class("az://mycontainer/")
    assert p2.blob == ""
    assert p2.container == "mycontainer"


@pytest.mark.parametrize("client_class", [AzureBlobClient, LocalAzureBlobClient])
def test_azureblobpath_nocreds(client_class, monkeypatch):
    monkeypatch.delenv("AZURE_STORAGE_CONNECTION_STRING", raising=False)
    with pytest.raises(MissingCredentialsError):
        client_class()



import pytest


def test_joins(rig):
    assert rig.create_cloud_path("a/b/c/d").name == "d"
    assert rig.create_cloud_path("a/b/c/d.file").name == "d.file"
    assert rig.create_cloud_path("a/b/c/d.file").stem == "d"
    assert rig.create_cloud_path("a/b/c/d.file").suffix == ".file"
    assert rig.create_cloud_path("a/b/c/d.tar.gz").suffixes == [".tar", ".gz"]
    assert (
        str(rig.create_cloud_path("a/b/c/d.file").with_suffix(".png"))
        == f"{rig.cloud_prefix}{rig.drive}/{rig.test_dir}/a/b/c/d.png"
    )

    assert rig.create_cloud_path("a") / "b" == rig.create_cloud_path("a/b")
    assert rig.create_cloud_path("a/b/c/d") / "../../b" == rig.create_cloud_path("a/b/b")

    assert rig.create_cloud_path("a/b/c/d").match("**/c/*")
    assert not rig.create_cloud_path("a/b/c/d").match("**/c")
    assert rig.create_cloud_path("a/b/c/d").match("a/*/c/d")

    assert rig.create_cloud_path("a/b/c/d").anchor == rig.cloud_prefix
    assert rig.create_cloud_path("a/b/c/d").parent == rig.create_cloud_path("a/b/c")

    assert rig.create_cloud_path("a/b/c/d").parents == [
        rig.create_cloud_path("a/b/c"),
        rig.create_cloud_path("a/b"),
        rig.create_cloud_path("a"),
        rig.path_class(f"{rig.cloud_prefix}{rig.drive}/{rig.test_dir}"),
        rig.path_class(f"{rig.cloud_prefix}{rig.drive}"),
    ]

    assert rig.create_cloud_path("a").joinpath("b", "c") == rig.create_cloud_path("a/b/c")

    assert rig.create_cloud_path("a/b/c").samefile(rig.create_cloud_path("a/b/c"))

    assert (
        rig.create_cloud_path("a/b/c").as_uri()
        == f"{rig.cloud_prefix}{rig.drive}/{rig.test_dir}/a/b/c"
    )

    assert rig.create_cloud_path("a/b/c/d").parts == (
        rig.cloud_prefix,
        rig.drive,
        rig.test_dir,
        "a",
        "b",
        "c",
        "d",
    )


def test_equality(rig):
    assert rig.create_cloud_path("a/b/foo") == rig.create_cloud_path("a/b/foo")
    assert hash(rig.create_cloud_path("a/b/foo")) == hash(rig.create_cloud_path("a/b/foo"))

    assert rig.create_cloud_path("a/b/foo") != rig.create_cloud_path("a/b/bar")
    assert hash(rig.create_cloud_path("a/b/foo")) != hash(rig.create_cloud_path("a/b/bar"))

    cp = rig.create_cloud_path("a/b/foo")
    assert cp != str(cp)
    assert cp != repr(cp)
    assert hash(cp) != hash(str(cp))
    assert hash(cp) != hash(repr(cp))


def test_sorting(rig):
    cp1 = rig.create_cloud_path("a/b/c")
    cp2 = rig.create_cloud_path("a/c/b")
    assert cp1 < cp2
    assert cp1 <= cp2
    assert not cp1 > cp2
    assert not cp1 >= cp2

    assert cp2 > cp1
    assert cp2 >= cp1
    assert not cp2 < cp1
    assert not cp2 <= cp1

    assert rig.create_cloud_path("a/b/c") <= rig.create_cloud_path("a/b/c")
    assert rig.create_cloud_path("a/b/c") >= rig.create_cloud_path("a/b/c")

    assert sorted(
        [
            rig.create_cloud_path("a/c/b"),
            rig.create_cloud_path("a/b/c"),
            rig.create_cloud_path("d/e/f"),
        ]
    ) == [
        rig.create_cloud_path("a/b/c"),
        rig.create_cloud_path("a/c/b"),
        rig.create_cloud_path("d/e/f"),
    ]

    with pytest.raises(TypeError):
        assert cp1 < str(cp1)
    with pytest.raises(TypeError):
        assert cp1 <= str(cp1)
    with pytest.raises(TypeError):
        assert cp1 > str(cp1)
    with pytest.raises(TypeError):
        assert cp1 >= str(cp1)



from cloudpathlib import CloudPath


def test_default_client_instantiation(rig):
    if not getattr(rig, "is_custom_s3", False):
        # Skip resetting the default client for custom S3 endpoint, but keep the other tests,
        # since they're still useful.
        rig.client_class._default_client = None

    # CloudPath dispatch
    p = CloudPath(f"{rig.cloud_prefix}{rig.drive}/{rig.test_dir}/dir_0/file0_0.txt")
    # Explicit class
    p2 = rig.create_cloud_path("dir_0/file0_0.txt")
    # Default client CloudPath constructor
    p3 = rig.client_class.get_default_client().CloudPath(
        f"{rig.cloud_prefix}{rig.drive}/{rig.test_dir}/dir_0/file0_0.txt"
    )
    # Default client path-class-name constructor
    p4 = getattr(rig.client_class.get_default_client(), rig.path_class.__name__)(
        f"{rig.cloud_prefix}{rig.drive}/{rig.test_dir}/dir_0/file0_0.txt"
    )

    # Check that paths are the same
    assert p == p2 == p3 == p4

    # Check that client is the same instance
    assert p.client is p2.client is p3.client is p4.client

    # Check the file content is the same
    assert p.read_bytes() == p2.read_bytes() == p3.read_bytes() == p4.read_bytes()

    # should be using same instance of client, so cache should be the same
    assert p._local == p2._local == p3._local == p4._local


def test_different_clients(rig):
    p = rig.create_cloud_path("dir_0/file0_0.txt")

    new_client = rig.client_class()
    p2 = new_client.CloudPath(f"{rig.cloud_prefix}{rig.drive}/{rig.test_dir}/dir_0/file0_0.txt")

    assert p.client is not p2.client
    assert p._local is not p2._local



import pytest

from inspect import signature

from cloudpathlib import AzureBlobClient, AzureBlobPath, GSClient, GSPath, S3Client, S3Path
from cloudpathlib.local import (
    LocalAzureBlobClient,
    LocalAzureBlobPath,
    LocalGSClient,
    LocalGSPath,
    LocalS3Client,
    LocalS3Path,
)


@pytest.mark.parametrize(
    "cloud_class,local_class",
    [
        (AzureBlobClient, LocalAzureBlobClient),
        (AzureBlobPath, LocalAzureBlobPath),
        (GSClient, LocalGSClient),
        (GSPath, LocalGSPath),
        (S3Client, LocalS3Client),
        (S3Path, LocalS3Path),
    ],
)
def test_interface(cloud_class, local_class):
    """Test that local class implements associated cloud class's interface"""

    cloud_attr_names = [attr for attr in dir(cloud_class) if not attr.startswith("_")]
    local_attr_names = [attr for attr in dir(local_class) if not attr.startswith("_")]

    assert set(cloud_attr_names).issubset(local_attr_names)

    for attr_name in cloud_attr_names:
        cloud_attr = getattr(cloud_class, attr_name)
        local_attr = getattr(local_class, attr_name)

        assert type(cloud_attr) == type(local_attr)
        if callable(cloud_attr):
            assert signature(cloud_attr).parameters == signature(local_attr).parameters


@pytest.mark.parametrize("client_class", [LocalAzureBlobClient, LocalGSClient, LocalS3Client])
def test_default_storage_dir(client_class, monkeypatch):
    """Test that local file storage for a LocalClient persists across client instantiations."""

    if client_class is LocalAzureBlobClient:
        monkeypatch.setenv("AZURE_STORAGE_CONNECTION_STRING", "")

    cloud_prefix = client_class._cloud_meta.path_class.cloud_prefix

    p1 = client_class().CloudPath(f"{cloud_prefix}drive/file.txt")
    p2 = client_class().CloudPath(f"{cloud_prefix}drive/file.txt")

    assert not p1.exists()
    assert not p2.exists()

    p1.write_text("hello")
    assert p1.exists()
    assert p1.read_text() == "hello"
    assert p2.exists()
    assert p2.read_text() == "hello"

    # clean up
    client_class.reset_default_storage_dir()


@pytest.mark.parametrize("client_class", [LocalAzureBlobClient, LocalGSClient, LocalS3Client])
def test_reset_default_storage_dir(client_class, monkeypatch):
    """Test that LocalClient default storage reset changes the default temp directory."""

    if client_class is LocalAzureBlobClient:
        monkeypatch.setenv("AZURE_STORAGE_CONNECTION_STRING", "")

    cloud_prefix = client_class._cloud_meta.path_class.cloud_prefix

    p1 = client_class().CloudPath(f"{cloud_prefix}drive/file.txt")
    client_class.reset_default_storage_dir()
    p2 = client_class().CloudPath(f"{cloud_prefix}drive/file.txt")

    assert not p1.exists()
    assert not p2.exists()

    p1.write_text("hello")
    assert p1.exists()
    assert p1.read_text() == "hello"
    assert not p2.exists()

    # clean up
    client_class.reset_default_storage_dir()



from concurrent.futures import ProcessPoolExecutor
from time import sleep

import pytest

from boto3.s3.transfer import TransferConfig
import botocore
from cloudpathlib import S3Client, S3Path
from cloudpathlib.local import LocalS3Path
import psutil


@pytest.mark.parametrize("path_class", [S3Path, LocalS3Path])
def test_s3path_properties(path_class):
    p = path_class("s3://bucket")
    assert p.key == ""
    assert p.bucket == "bucket"

    p2 = path_class("s3://bucket/")
    assert p2.key == ""
    assert p2.bucket == "bucket"


def test_transfer_config(s3_rig, tmp_path):
    transfer_config = TransferConfig(multipart_threshold=50)
    client = s3_rig.client_class(boto3_transfer_config=transfer_config)
    assert client.boto3_transfer_config.multipart_threshold == 50
    # check defaults are inherited as well
    assert client.boto3_transfer_config.use_threads

    # download
    client.set_as_default_client()
    p = s3_rig.create_cloud_path("dir_0/file0_0.txt")

    dl_dir = tmp_path
    assert not (dl_dir / p.name).exists()
    p.download_to(dl_dir)

    # we can only check the configs are actually passed on the mock
    if not s3_rig.live_server:
        assert client.s3.download_config == transfer_config

    # upload
    p2 = s3_rig.create_cloud_path("dir_0/file0_0_uploaded.txt")
    assert not p2.exists()
    p2.upload_from(dl_dir / p.name)

    # we can only check the configs are actually passed on the mock
    if not s3_rig.live_server:
        assert client.s3.upload_config == transfer_config

    p2.unlink()


def _download_with_threads(s3_rig, tmp_path, use_threads):
    """Job used by tests to ensure Transfer config changes are
    actually passed through to boto3 and respected.
    """
    sleep(1)  # give test monitoring process time to start watching

    transfer_config = TransferConfig(
        max_concurrency=100,
        use_threads=use_threads,
        multipart_chunksize=1 * 1024,
        multipart_threshold=10 * 1024,
    )
    client = s3_rig.client_class(boto3_transfer_config=transfer_config)
    p = client.CloudPath(f"s3://{s3_rig.drive}/{s3_rig.test_dir}/dir_0/file0_to_download.txt")

    assert not p.exists()

    # file should be about 60KB
    text = "lalala" * 10_000
    p.write_text(text)

    assert p.exists()

    # assert not (dl_dir / p.name).exists()
    p.download_to(tmp_path)

    p.unlink()

    assert not p.exists()


def test_transfer_config_live(s3_rig, tmp_path):
    """Tests that boto3 receives and respects the transfer config
    when working with a live backend. Does this by observing
    if the `use_threads` parameter changes to number of threads
    used by a child process that does a download.
    """
    if not s3_rig.live_server:
        pytest.skip("This test only runs against live servers.")

    def _execute_on_subprocess_and_observe(use_threads):
        main_test_process = psutil.Process().pid

        with ProcessPoolExecutor(max_workers=1) as executor:
            job = executor.submit(
                _download_with_threads,
                s3_rig=s3_rig,
                tmp_path=tmp_path,
                use_threads=use_threads,
            )

            max_threads = 0

            # timeout after 100 seconds
            for _ in range(1000):
                worker_process_id = (
                    psutil.Process(main_test_process).children()[-1].pid
                )  # most recently started child
                n_thread = psutil.Process(worker_process_id).num_threads()

                # observe number of threads used
                max_threads = max(max_threads, n_thread)

                sleep(0.1)

                if job.done():
                    _ = job.result()  # raises if job raised
                    break

            return max_threads

    # usually ~3 threads are spun up whe use_threads is False
    assert _execute_on_subprocess_and_observe(use_threads=False) < 5

    # usually ~15 threads are spun up whe use_threads is True
    assert _execute_on_subprocess_and_observe(use_threads=True) > 10


def test_fake_directories(s3_like_rig):
    """S3 can have "fake" directories created
    either in the AWS S3 Console or by uploading
    a 0 size object ending in a `/`. If these objects
    exist, we want to treat them as directories.

    Note: Our normal tests do _not_ create folders in this
    way, so this test is the only one to exercise these "fake" dirs.

    Ref: https://github.com/boto/boto3/issues/377
    """
    if not s3_like_rig.live_server:
        pytest.skip("This test only runs against live servers.")

    boto3_s3_client = s3_like_rig.client_class._default_client.client

    response = boto3_s3_client.put_object(
        Bucket=f"{s3_like_rig.drive}",
        Body="",
        Key=f"{s3_like_rig.test_dir}/fake_directory/",
    )

    assert response["ResponseMetadata"]["HTTPStatusCode"] == 200

    # test either way of referring to the directory (with and w/o terminal slash)
    fake_dir_slash = s3_like_rig.create_cloud_path("fake_directory/")
    fake_dir_no_slash = s3_like_rig.create_cloud_path("fake_directory")

    for test_case in [fake_dir_no_slash, fake_dir_slash]:
        assert test_case.exists()
        assert test_case.is_dir()
        assert not test_case.is_file()


def test_no_sign_request(s3_rig, tmp_path):
    """Tests that we can pass no_sign_request to the S3Client and we will
    be able to access public resources but not private ones.
    """
    if not s3_rig.live_server:
        pytest.skip("This test only runs against live servers.")

    client = s3_rig.client_class(no_sign_request=True)

    # unsigned can access public path (part of AWS open data)
    p = client.CloudPath(
        "s3://ladi/Images/FEMA_CAP/2020/70349/DSC_0001_5a63d42e-27c6-448a-84f1-bfc632125b8e.jpg"
    )
    assert p.exists()

    p.download_to(tmp_path)
    assert (tmp_path / p.name).read_bytes() == p.read_bytes()

    # unsigned raises for private S3 file that exists
    p = client.CloudPath(f"s3://{s3_rig.drive}/dir_0/file0_to_download.txt")
    with pytest.raises(botocore.exceptions.ClientError):
        p.exists()


def test_aws_endpoint_url_env(monkeypatch):
    """Allows setting endpoint_url from env variable
    until upstream boto3 PR is merged.
    https://github.com/boto/boto3/pull/2746
    """
    s3_url = "https://s3.amazonaws.com"
    localstack_url = "http://localhost:4566"

    s3_client = S3Client()
    assert s3_client.client.meta.endpoint_url == s3_url

    monkeypatch.setenv("AWS_ENDPOINT_URL", localstack_url)
    s3_client_custom_endpoint = S3Client()
    assert s3_client_custom_endpoint.client.meta.endpoint_url == localstack_url



import os

import pytest

from cloudpathlib import AzureBlobPath, CloudPath, GSPath, S3Path
from cloudpathlib.exceptions import InvalidPrefixError, MissingDependenciesError


@pytest.mark.parametrize(
    "path_class, cloud_path",
    [
        (AzureBlobPath, "az://b/k"),
        (AzureBlobPath, "AZ://b/k"),
        (AzureBlobPath, "Az://b/k"),
        (AzureBlobPath, "aZ://b/k"),
        (S3Path, "s3://b/k"),
        (S3Path, "S3://b/k"),
        (GSPath, "gs://b/k"),
        (GSPath, "GS://b/k"),
        (GSPath, "Gs://b/k"),
        (GSPath, "gS://b/k"),
    ],
)
def test_dispatch(path_class, cloud_path, monkeypatch):
    """Test that CloudPath(...) appropriately dispatches to the correct cloud's implementation
    class.
    """
    if path_class == AzureBlobPath:
        monkeypatch.setenv("AZURE_STORAGE_CONNECTION_STRING", "AccountName=fake;AccountKey=fake2;")

    assert isinstance(CloudPath(cloud_path), path_class)


def test_dispatch_error():
    with pytest.raises(InvalidPrefixError):
        CloudPath("pp://b/k")


@pytest.mark.parametrize("path", ["b/k", "b/k", "b/k.file", "b/k", "b"])
def test_instantiation(rig, path):
    # check two cases of prefix
    for prefix in [rig.cloud_prefix.lower(), rig.cloud_prefix.upper()]:
        expected = prefix + path
        p = rig.path_class(expected)
        assert repr(p) == f"{rig.path_class.__name__}('{expected}')"
        assert str(p) == expected

        assert p._no_prefix == expected.split("://", 1)[-1]

        assert p._url.scheme == expected.split("://", 1)[0].lower()
        assert p._url.netloc == expected.split("://", 1)[-1].split("/")[0]

        assert str(p._path) == expected.split(":/", 1)[-1]


def test_instantiation_errors(rig):
    with pytest.raises(TypeError):
        rig.path_class()

    with pytest.raises(InvalidPrefixError):
        rig.path_class("NOT_S3_PATH")


def test_idempotency(rig):
    rig.client_class._default_client = None

    client = rig.client_class()
    p = client.CloudPath(f"{rig.cloud_prefix}{rig.drive}/{rig.test_dir}/dir_0/file0_0.txt")

    p2 = CloudPath(p)
    assert p == p2
    assert p.client == p2.client


def test_dependencies_not_loaded(rig, monkeypatch):
    monkeypatch.setattr(rig.path_class._cloud_meta, "dependencies_loaded", False)
    with pytest.raises(MissingDependenciesError):
        CloudPath(f"{rig.cloud_prefix}{rig.drive}/{rig.test_dir}/dir_0/file0_0.txt")
    with pytest.raises(MissingDependenciesError):
        rig.create_cloud_path("dir_0/file0_0.txt")


def test_is_pathlike(rig):
    p = rig.create_cloud_path("dir_0")
    assert isinstance(p, os.PathLike)



from pathlib import Path

from pydantic import BaseModel, ValidationError
import pytest

from cloudpathlib.anypath import AnyPath


def test_pydantic_cloudpath(rig):
    class PydanticModel(BaseModel):
        cloud_path: rig.path_class

    cp = rig.create_cloud_path("a/b/c")

    obj = PydanticModel(cloud_path=cp)
    assert obj.cloud_path == cp

    obj = PydanticModel(cloud_path=str(cp))
    assert obj.cloud_path == cp

    with pytest.raises(ValidationError):
        _ = PydanticModel(cloud_path=0)


def test_pydantic_anypath(rig):
    class PydanticModel(BaseModel):
        any_path: AnyPath

    cp = rig.create_cloud_path("a/b/c")

    obj = PydanticModel(any_path=cp)
    assert obj.any_path == cp

    obj = PydanticModel(any_path=str(cp))
    assert obj.any_path == cp

    obj = PydanticModel(any_path=Path("a/b/c"))
    assert obj.any_path == Path("a/b/c")

    obj = PydanticModel(any_path="a/b/c")
    assert obj.any_path == Path("a/b/c")

    with pytest.raises(ValidationError):
        obj = PydanticModel(any_path=0)



from datetime import datetime
import os
from pathlib import PurePosixPath
from time import sleep

import pytest

from cloudpathlib.exceptions import CloudPathIsADirectoryError, DirectoryNotEmptyError


def test_file_discovery(rig):
    p = rig.create_cloud_path("dir_0/file0_0.txt")
    assert p.exists()

    p2 = rig.create_cloud_path("dir_0/not_a_file")
    assert not p2.exists()
    p2.touch()
    assert p2.exists()
    p2.unlink()

    p3 = rig.create_cloud_path("dir_0/")
    assert p3.exists()
    assert len(list(p3.iterdir())) == 3
    assert len(list(p3.glob("**/*"))) == 3

    with pytest.raises(CloudPathIsADirectoryError):
        p3.unlink()

    with pytest.raises(DirectoryNotEmptyError):
        p3.rmdir()
    p3.rmtree()
    assert not p3.exists()

    p4 = rig.create_cloud_path("")
    assert p4.exists()

    assert len(list(p4.iterdir())) == 1  # only bucket/dir_1/ should still exist
    assert len(list(p4.glob("**/*"))) == 4

    assert list(p4.glob("**/*")) == list(p4.rglob("*"))

    # try rmtree on dir_1, which has a nested directory
    p5 = rig.create_cloud_path("dir_1/")
    assert p5.exists()
    assert len(list(p5.glob("**/*"))) == 3
    with pytest.raises(CloudPathIsADirectoryError):
        p5.unlink()

    with pytest.raises(DirectoryNotEmptyError):
        p5.rmdir()

    p5.rmtree()
    assert not p5.exists()


def test_is_dir_is_file(rig, tmp_path):
    # test on directories
    dir_slash = rig.create_cloud_path("dir_0/")
    dir_no_slash = rig.create_cloud_path("dir_0")
    dir_nested_slash = rig.create_cloud_path("dir_1/dir_1_0/")
    dir_nested_no_slash = rig.create_cloud_path("dir_1/dir_1_0")

    for test_case in [dir_slash, dir_no_slash, dir_nested_slash, dir_nested_no_slash]:
        assert test_case.is_dir()
        assert not test_case.is_file()

    file = rig.create_cloud_path("dir_0/file0_0.txt")
    file_nested = rig.create_cloud_path("dir_1/dir_1_0/file_1_0_0.txt")

    for test_case in [file, file_nested]:
        assert test_case.is_file()
        assert not test_case.is_dir()

    # does not exist (same behavior as pathlib.Path that does not exist)
    non_existant = rig.create_cloud_path("dir_0/not_a_file")
    assert not non_existant.is_file()
    assert not non_existant.is_dir()


def test_file_read_writes(rig, tmp_path):
    p = rig.create_cloud_path("dir_0/file0_0.txt")
    p2 = rig.create_cloud_path("dir_0/not_a_file")
    p3 = rig.create_cloud_path("")

    text = "lalala" * 10_000

    p.write_text(text)
    assert p.read_text() == text
    p2.write_text(text)

    # sleep between writes to p to ensure different
    # modified times
    sleep(1)

    p.write_bytes(p2.read_bytes())
    assert p.read_text() == p2.read_text()

    before_touch = datetime.now()
    sleep(1)
    p.touch()
    if not getattr(rig, "is_custom_s3", False):
        # Our S3Path.touch implementation does not update mod time for MinIO
        assert datetime.fromtimestamp(p.stat().st_mtime) > before_touch

    # no-op
    p.mkdir()

    assert p.etag is not None

    dest = rig.create_cloud_path("dir2/new_file0_0.txt")
    assert not dest.exists()
    p.rename(dest)
    assert dest.exists()

    assert not p.exists()
    p.touch()
    dest.replace(p)
    assert p.exists()

    dl_file = tmp_path / "file"
    p.download_to(dl_file)
    assert dl_file.exists()
    assert p.read_text() == dl_file.read_text()

    dl_dir = tmp_path / "directory"
    dl_dir.mkdir(parents=True, exist_ok=True)
    p3.download_to(dl_dir)
    cloud_rel_paths = sorted(
        # CloudPath("prefix://drive/dir/file.txt")._no_prefix_no_drive = "/dir/file.txt"
        [p._no_prefix_no_drive[len(rig.test_dir) + 2 :] for p in p3.glob("**/*")]
    )
    dled_rel_paths = sorted(
        [str(PurePosixPath(p.relative_to(dl_dir))) for p in dl_dir.glob("**/*")]
    )
    assert cloud_rel_paths == dled_rel_paths


def test_cloud_path_download_to(rig, tmp_path):
    p = rig.create_cloud_path("dir_0/file0_0.txt")
    dl_dir = tmp_path
    assert not (dl_dir / p.name).exists()
    p.download_to(dl_dir)
    assert (dl_dir / p.name).is_file()


def test_fspath(rig):
    p = rig.create_cloud_path("dir_0")
    assert os.fspath(p) == p.fspath


def test_os_open(rig):
    p = rig.create_cloud_path("dir_0/file0_0.txt")
    with open(p, "r") as f:
        assert f.readable()
